package bench

import bench.BenchRastriginAverageNSGAII.replication
import freedsl.dsl._
import mgo._
import mgo.breeding._
import mgo.algorithm._
import mgo.contexts._
import mgo.test._
import better.files._
import cats.data._
import cats.implicits._
import mgo.stop.StopCondition

object BenchRastrigin {
  def headers(genomeSize: Int) =
    "evaluations" :: List.tabulate(genomeSize)(e => s"g$e") ::: "rastrigin" :: "fitness" :: Nil

  val mu = 50
  val lambda = 50
  val maxFitnessEvaluation = 50000

  def stop[M[_]: cats.Monad: Generation, I](f: Long => Boolean): StopCondition[M, I] =
    Kleisli { (individuals: Vector[I]) =>
      for {
        cg <- Generation[M].get
      } yield f(cg)
    }

}

object BenchRastriginAverageNSGAII extends App {

  import nsga2._

  val directory = File("/tmp/rastrigin_avg/")
  directory.delete(swallowIOExceptions = true)
  directory.createDirectories()

  val numberOfReplications = 100

  def numberOfEvaluation(g: Long) =
    (BenchRastrigin.mu * numberOfReplications) + g * numberOfReplications * BenchRastrigin.lambda

  def replication(r: Int) = {
    val rng = new util.Random(r + 20000)

    def evolution[M[_] : Generation : Random : cats.Monad : StartTime : IO] = {
      val (nsga2, source) = sourceOf {
        def fitness(x: Vector[Double]) = {
          val samples = Vector.fill(numberOfReplications)(additiveGaussianNoise(0.25, rastrigin.compute).apply(rng, x).head)
          Vector(average(samples))
        }

        NSGA2[M](
          mu = BenchRastrigin.mu,
          lambda = BenchRastrigin.lambda,
          fitness = fitness(_),
          genomeSize = 10,
          operators = ManualOperators[M](sbxC(0.5), bga(mutationRate = _ => 0.5, mutationRange = 0.1))
        )
      }

      def save(generation: Long, population: Vector[Individual]) = {
        val lines = result(population, dropWave.scale).map { case (g, v) => g ++ v }.map(_.mkString(","))
        //(directory / s"result.csv") overwrite lines.mkString("\n")
        val bestIndividual = population.minBy(_.fitness.head)
        val genome = bestIndividual.genome.values
        val bestFitness = bestIndividual.fitness.head
        val scaledGenome = rastrigin.scale(genome.toVector)

        (directory / s"best$r.csv") appendLine s"${numberOfEvaluation(generation)}, ${scaledGenome.mkString(",")}, ${rastrigin.compute(genome.toVector)}, ${bestFitness}"
      }

      //(directory / "source") overwrite source

      nsga2.
        until(BenchRastrigin.stop(g => numberOfEvaluation(g) >= 50000)).
        trace((s, is) => save(s.generation, is)).
        evolution
    }

    NSGA2.run(new util.Random(r + 1000)) { impl =>
      import impl._
      (directory / s"best$r.csv") appendLine BenchRastrigin.headers(10).mkString(", ")
      evolution[DSL].eval
    }
  }

  for {
    r <- 0 until 100 par
  } {
    replication(r)
  }

}

object BenchRastriginNoisyNSGAII extends App {

  import noisynsga2._

  val directory = File("/tmp/rastrigin_noisy/")
  directory.delete(swallowIOExceptions = true)
  directory.createDirectories()

  def numberOfEvaluation(g: Long) = BenchRastrigin.mu + (g * BenchRastrigin.lambda)

  def replication(r: Int) = {
    def evolution[M[_] : Generation : Random : cats.Monad : StartTime : IO] = {
      val (nsga2, source) = sourceOf {
        NoisyNSGA2[M](
          mu = BenchRastrigin.mu,
          lambda = BenchRastrigin.lambda,
          fitness = additiveGaussianNoise(0.25, rastrigin.compute),
          genomeSize = 10,
          aggregation = averageAggregation(_),
          operators = ManualOperators[M](sbxC(0.5), bga(mutationRate = _ => 0.5, mutationRange = 0.1))
        )
      }

      def save(generation: Long, population: Vector[Individual]) = {
//        val lines = {
//          val res = population.map { i =>
//            dropWave.scale(i.genome.values.toVector) ++ i.fitnessHistory.transpose.head
//          }
//          res.map(_.mkString(","))
//        }

        //(directory / s"$generation.csv") overwrite lines.mkString("\n")

        def best = {
          val full =
            population.map { i =>
              i.genome.values -> i.fitnessHistory.transpose.head
            }.filter(_._2.size == 100)

          full.isEmpty match {
            case true => None
            case false =>
              Some(full.minBy(v => average(v._2.toVector)))
          }
        }

        best foreach  {
          case (bestGenome, bestFitness) =>
            (directory / s"best$r.csv") appendLine s"${BenchRastrigin.mu + (generation * BenchRastrigin.lambda)}, ${rastrigin.scale(bestGenome.toVector).mkString(",")}, ${rastrigin.compute(bestGenome.toVector)}, ${average(bestFitness)}"
        }

      }

      //(directory / "source") overwrite source

      nsga2.
        until(BenchRastrigin.stop(g => numberOfEvaluation(g) >= 50000)).
        trace((s, is) => save(s.generation, is)).
        evolution
    }

    NoisyNSGA2.run(new util.Random(1000 + r)) { impl =>
      import impl._
      (directory / s"best$r.csv") appendLine BenchRastrigin.headers(10).mkString(", ")
      evolution[DSL].eval
    }

  }

  for {
    r <- 0 until 100 par
  } {
    replication(r)
  }

}